import asyncio
import logging

import pytest


@pytest.fixture
async def yield_fixture():
    logging.info("Setup")
    await asyncio.sleep(0)
    yield True
    await asyncio.sleep(0)
    logging.info("Teardown")



import asyncio
from functools import partial, total_ordering
from typing import Union

from .thread_pool import threaded


def proxy_method_async(name, in_executor=True):
    def wrap_to_future(loop, func, *args, **kwargs):
        future = loop.create_future()

        def _inner():
            try:
                return future.set_result(func(*args, **kwargs))
            except Exception as e:
                return future.set_exception(e)

        loop.call_soon(_inner)
        return future

    def wrap_to_thread(loop, func, executor, *args, **kwargs):
        callee = partial(func, *args, **kwargs)
        return loop.run_in_executor(executor, callee)

    async def method(self, *args, **kwargs):
        func = getattr(self.fp, name)

        if in_executor:
            return await wrap_to_thread(
                self.loop, func, self.executor, *args, **kwargs
            )

        return await wrap_to_future(self.loop, func, *args, **kwargs)

    method.__name__ = name
    return method


def proxy_method(name):
    def method(self, *args, **kwargs):
        return getattr(self.fp, name)(*args, **kwargs)

    method.__name__ = name
    return method


def proxy_property(name):
    def fset(self, value):
        setattr(self.fp, name, value)

    def fget(self):
        return getattr(self.fp, name)

    def fdel(self):
        delattr(self.fp, name)

    return property(fget, fset, fdel)


@total_ordering
class AsyncFileIOBase:
    __slots__ = ('loop', '__opener', 'fp', 'executor', '__iterator_lock')

    opener = staticmethod(threaded(open))

    def __init__(self, fname, mode="r", executor=None, *args, **kwargs):
        self.loop = kwargs.pop('loop', asyncio.get_event_loop())
        self.fp = None
        self.executor = executor
        self.__opener = partial(self.opener, fname, mode, *args, **kwargs)
        self.__iterator_lock = asyncio.Lock(loop=self.loop)

    def closed(self):
        return self.fp.closed

    async def open(self):
        if self.fp is not None:
            return

        self.fp = await self.__opener()

    def __await__(self):
        yield from self.open().__await__()
        return self

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.loop.run_in_executor(
            None, self.fp.__exit__, exc_type, exc_val, exc_tb
        )

    def __del__(self):
        if not self.fp or self.fp.closed:
            return

        self.fp.close()
        del self.fp

    def __aiter__(self):
        return self

    async def __anext__(self):
        async with self.__iterator_lock:
            line = await self.readline()

        if not len(line):
            raise StopAsyncIteration

        return line

    def __eq__(self, other: "async_open"):
        return (
            self.__class__, self.fp.__eq__(other)
        ) == (
            other.__class__, self.fp.__eq__(other)
        )

    def __lt__(self, other: "async_open"):
        return self.fp < other.fp

    def __hash__(self):
        return hash((self.__class__, self.fp))

    fileno = proxy_method('fileno')
    isatty = proxy_method('isatty')

    mode = proxy_property('mode')
    name = proxy_property('name')

    close = proxy_method_async('close')
    detach = proxy_method_async('detach')
    flush = proxy_method_async('flush')
    peek = proxy_method_async('peek')
    raw = proxy_method_async('raw')
    read = proxy_method_async('read')
    read1 = proxy_method_async('read1')
    readinto = proxy_method_async('readinto')
    readinto1 = proxy_method_async('readinto1')
    readline = proxy_method_async('readline')
    readlines = proxy_method_async('readlines')
    seek = proxy_method_async('seek')
    peek = proxy_method_async('peek')
    truncate = proxy_method_async('truncate')
    write = proxy_method_async('write')
    writelines = proxy_method_async('writelines')

    tell = proxy_method_async('tell', in_executor=False)
    readable = proxy_method_async('readable', in_executor=False)
    seekable = proxy_method_async('seekable', in_executor=False)
    writable = proxy_method_async('writable', in_executor=False)


class AsyncTextFileIOBase:
    newlines = proxy_property('newlines')
    errors = proxy_property('errors')
    line_buffering = proxy_property('line_buffering')
    encoding = proxy_property('encoding')
    buffer = proxy_property('buffer')


class AsyncBytesFileIOBase:
    raw = proxy_property('raw')


class AsyncTextFileIO(AsyncFileIOBase, AsyncTextFileIOBase):
    pass


class AsyncBytesFileIO(AsyncFileIOBase, AsyncBytesFileIOBase):
    pass


AsyncFileT = Union[
    AsyncBytesFileIO,
    AsyncTextFileIO,
]


def async_open(fname, mode="r", *args, **kwargs) -> AsyncFileT:
    if 'b' in mode:
        return AsyncBytesFileIO(fname, mode=mode, *args, **kwargs)
    return AsyncTextFileIO(fname, mode=mode, *args, **kwargs)



import inspect


class Signal:

    __slots__ = ('_receivers', '_is_frozen')

    def __init__(self):
        self._receivers = set()

    def connect(self, receiver):
        if self.is_frozen:
            raise RuntimeError(
                "Can't connect receiver (%r) to the frozen signal",
                receiver,
            )

        if not inspect.iscoroutinefunction(receiver):
            raise RuntimeError('%r is not a coroutine function', receiver)

        self._receivers.add(receiver)

    async def call(self, *args, **kwargs):
        for receiver in self._receivers:
            await receiver(*args, **kwargs)

    def copy(self):
        clone = Signal()
        # unfreeze on copy
        clone._receivers = set(self._receivers)
        return clone

    @property
    def is_frozen(self):
        return isinstance(self._receivers, frozenset)

    def freeze(self):
        self._receivers = frozenset(self._receivers)


def receiver(s: Signal):
    def decorator(func):
        s.connect(func)
        return func

    return decorator



import asyncio
import inspect
import time
import typing
from collections import deque

T = typing.TypeVar('T')
R = typing.TypeVar('R')

GenType = typing.Generator[T, R, None]
FuncType = typing.Callable[[], GenType]


class IteratorWrapper(typing.AsyncIterator):
    __slots__ = (
        "loop", "__closed", "executor", "__close_event",
        "__queue", "__queue_maxsize", "__gen_task", "__gen_func"
    )

    def __init__(self, gen_func: FuncType, loop=None,
                 max_size=0, executor=None):

        self.loop = loop or asyncio.get_event_loop()
        self.executor = executor

        self.__closed = False
        self.__close_event = asyncio.Event(loop=self.loop)
        self.__queue = deque()
        self.__queue_maxsize = max_size
        self.__gen_task = None      # type: asyncio.Task
        self.__gen_func = gen_func  # type: typing.Callable

    @property
    def closed(self):
        return self.__closed

    @staticmethod
    def __throw(_):
        pass

    def __in_thread(self):
        try:
            gen = iter(self.__gen_func())

            throw = self.__throw
            if inspect.isgenerator(gen):
                throw = gen.throw

            while not self.closed:
                item = next(gen)

                while len(self.__queue) > self.__queue_maxsize:
                    time.sleep(0)

                    if self.closed:
                        throw(asyncio.CancelledError())
                        return

                self.__queue.append((item, False))
        except StopIteration as e:
            if self.closed:
                return
            self.__queue.append((e, None))
        except Exception as e:
            if self.closed:
                return
            self.__queue.append((e, True))
        finally:
            self.loop.call_soon_threadsafe(self.__close_event.set)

    async def close(self):
        self.__closed = True
        self.__queue.clear()

        if not self.__gen_task.done():
            self.__gen_task.cancel()

        await self.__close_event.wait()
        await asyncio.gather(
            self.__gen_task, loop=self.loop, return_exceptions=True
        )
        del self.__queue

    def __aiter__(self):
        if self.__gen_task is not None:
            return self

        self.__gen_task = self.loop.run_in_executor(
            self.executor, self.__in_thread
        )
        return self

    async def __anext__(self) -> typing.Awaitable[T]:
        while not len(self.__queue):
            await asyncio.sleep(0, loop=self.loop)

        item, is_exc = self.__queue.popleft()

        if is_exc is None:
            await self.close()
            raise StopAsyncIteration(*item.args) from item
        elif is_exc:
            await self.close()
            raise item from item

        return item

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.closed:
            return

        await self.close()



import asyncio
import logging

import pytest
from async_generator import yield_, async_generator


@pytest.fixture
@async_generator
async def yield_fixture():
    logging.info("Setup")
    await asyncio.sleep(0)
    await yield_(True)
    await asyncio.sleep(0)
    logging.info("Teardown")


async def test_yield_fixture(yield_fixture):
    assert yield_fixture is True



import asyncio

from aiomisc.context import Context, get_context


class ServiceMeta(type):
    def __new__(cls, name, bases, namespace, **kwds):
        instance = type.__new__(cls, name, bases, dict(namespace))

        for key in ('__async_required__', '__required__'):
            setattr(instance, key, frozenset(getattr(instance, key, ())))

        check_instance = all(
            asyncio.iscoroutinefunction(getattr(instance, method))
            for method in instance.__async_required__
        )

        if not check_instance:
            raise TypeError(
                'Following methods must be coroutine functions',
                tuple('%s.%s' % (name, m) for m in instance.__async_required__)
            )

        return instance


class Service(metaclass=ServiceMeta):
    __async_required__ = 'start', 'stop'
    __required__ = ()

    def __init__(self, **kwargs):
        lost_kw = self.__required__ - kwargs.keys()
        if lost_kw:
            raise AttributeError('Absent attributes', lost_kw)

        self.loop = None
        self._set_params(**kwargs)
        self.__context = None
        self.start_event = None       # type: asyncio.Event

    @property
    def context(self) -> Context:
        if self.__context is None:
            self.__context = get_context()
        return self.__context

    def set_loop(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop
        self.start_event = asyncio.Event(loop=self.loop)

    def _set_params(self, **kwargs):
        for name, value in kwargs.items():
            setattr(self, name, value)

    async def start(self):
        raise NotImplementedError

    async def stop(self, exception: Exception = None):
        pass


class SimpleServer(Service):
    def __init__(self, **kwargs):
        self.server = None
        super().__init__(**kwargs)

    async def start(self):
        raise NotImplementedError

    async def stop(self, exc: Exception = None):
        self.server.close()
